#!/usr/bin/env python3
import sys
import torch

def read_file(label):
  with open(label, 'r', encoding='utf-8') as f:
      l1 = f.readlines()
      l1 = [line.strip('\n').split(' ') for line in l1]
      x1 = [line[2:] for line in l1]
      y1 = [line[0] for line in l1]
  return(x1,y1)

def evaluate(yp,yt):
  yp[yp >= 0.5] = 1
  yp[yp < 0.1] = -1
  corrects = torch.sum(torch.eq(yp,yt)).item()
  return(corrects)

def isBGCRatio(outputs):
  count = outputs[outputs[:, -1]<0.5].shape[0]
  return count/outputs.shape[0]

if(__name__ == '__main__'):
  label,nolabel,test = sys.argv[1],sys.argv[2],sys.argv[3]
  x1,y1,x2,x3 = read_file(label,nolabel,test)
  print(x1[0:2],y1[0:2])
  print(x2[0:2])
  print(x3[0:2])



import torch
from torch.utils.data import Dataset
import lmdb
import pandas as pd
import numpy as np
import random
import pickle


class DataReader:

    def __init__(self, datasetPath, test_ratio = 0.2) -> None:
        self.datasetPath = datasetPath
        self.test_ratio = test_ratio
        self.embedding_dim = 320
        self.__getData()
        self.__train_test_split()

    def __getData(self):
        self.df = pd.read_csv(self.datasetPath)
        # 读取csv并展开为list便于处理
        self.df['labels'] = self.df['labels'].apply(lambda x:x.split())
        self.df['TDsentence'] = self.df['TDsentence'].apply(lambda x:x.split())
        self.df['TDlabels'] = self.df['TDlabels'].apply(lambda x:x.split()).apply(lambda x:[int(eval(i)) for i in x])
        self.df['labels_rep'] = self.df['labels'].apply(lambda x:x[0]) # 仅取第一个便于分层抽样

        self.labels_list = ['Alkaloid', 'Terpene', 'NRP', 'Polyketide', 'RiPP', 'Saccharide', 'Other']
        print(self.labels_list)
        self.labels_num = len(self.labels_list)

    def __len__(self):
        return len(self.df)
    
    def __getitem__(self, idx):

        sentence = self.df['TDsentence'][idx]
        labels = self.df['labels'][idx]
        TDlabels = self.df['TDlabels'][idx]

        if labels == ['Unknown']:
            labels = random.choice(self.labels_list)
        # 在labels未知时随机一个值以便后面eval时能计算
        labels_onehot = [1 if label in labels else 0 for label in self.labels_list]
        return sentence, labels_onehot, TDlabels
    
    def __train_test_split(self):
        groups = self.df.groupby('labels_rep')
        self.train = []
        self.test = []
        # 根据代表的类别进行分层抽样
        for k, v in groups.groups.items():
            # k为代表的标签
            # v为索引
            vv = list(v)
            # print(vv)
            random.shuffle(vv)
            self.train += vv[int(v.size*self.test_ratio):]
            self.test += vv[:int(v.size*self.test_ratio)]
        # print(len(self.train))

    

    
class BGCLabelsDataset(Dataset):
    def __init__(self, data, lmdbPath, mode='train') -> None:
        super().__init__()
        self.data = data
        self.mode = mode
        self.env = lmdb.open(lmdbPath, readonly=True)
        self.txn = self.env.begin()

    def __len__(self):
        if self.mode == 'train':
            # print("length of training dataset",len(self.data.train))
            return len(self.data.train)
        elif self.mode == 'test':
            # print("length of testing dataset",len(self.data.test))
            return len(self.data.test)
        elif self.mode == 'eval':
            return len(self.data)
        else:
            raise ValueError("No such mode: {}!".format(self.mode))
    
    def __getitem__(self, idx):
        if self.mode == 'train':
            index = self.data.train[idx]
        elif self.mode == 'test':
            index = self.data.test[idx]
        elif self.mode == 'eval':
            index = idx
        else:
            raise ValueError("No such mode: {}!".format(self.mode))
        
        # print(index)
        sentence, labels_onehot, TDlabels = self.data[index]
        sentence_embedding = [pickle.loads(self.txn.get(str(word).encode('ascii')))['mean_representations'][6] for word in sentence]
        # print(sentence_embedding)
        sentence_embedding = torch.stack(sentence_embedding)
        labels_onehot = np.array(labels_onehot)
        TDlabels = np.array(TDlabels)
        # assert sentence_embedding.shape[0]==len(TDlabels), f'{idx} sample: length of sentence: {len(sentence_embedding)}, length of labels: {len(TDlabels)}'
        # print(index, self.data.df.iloc[index], sentence_embedding.shape, labels_onehot.shape, TDlabels.shape)
        return sentence_embedding, torch.tensor(labels_onehot, dtype=torch.float32), torch.tensor(TDlabels, dtype=torch.float32)
    
    
    def __del__(self):
        self.env.close()
        # super().__del__()
        # print('env closed')





        
            
            







import pandas as pd
import argparse
from tqdm import tqdm
from multiprocessing import Pool
import os
from pathlib import Path

from .baseCommand import baseCommand

class splitCommand(baseCommand):
    name = "split"
    description = "Split genomes into sequences which have 128 genes"

    def add_arguments(self, parser):
        parser.add_argument('--genomesPath', type=Path, required=True, help='genomes csv path')
        parser.add_argument('--outputPath', type=Path, required=False, default=Path('./output/'), help='output path')
        parser.add_argument('--name', type=str, required=False, default='split', help='name of the output file')
        parser.add_argument('--threads', type=int, required=False, default=10, help='number of cpu threads to split genomes')

    def handle(self, args):
        split = splitModule(args)
        split.split_genomes()

def split_genome(input):
    index, series = input
    genome = series['Genome']
    gene_seq = eval(series['Gene_sequence'])
    dataFrame = pd.DataFrame(columns=['ID', 'sentence', 'labels', 'isBGC', 'TDsentence', 'TDlabels'])
    if len(gene_seq) <128:
        while len(gene_seq) < 128:
            gene_seq += gene_seq
        sentence = gene_seq[:128]
        ID = genome + '_' + str(0)
        dataFrame.loc[len(dataFrame)] = [ID, None, 'Unknown', 'No', sentence, [0]*2]
        return dataFrame
    for i in range(0, len(gene_seq)//128+1):
        if (i+1)*128 > len(gene_seq):
            sentence = gene_seq[-128:]
        else:
            sentence = gene_seq[i*128:(i+1)*128]
        ID = genome + '_' + str(i)
        dataFrame.loc[len(dataFrame)] = [ID, None, 'Unknown', 'No', sentence, [0]*2]
        # len(gene_seq) == n*128, deduplicate
        if (i+1)*128 == len(gene_seq):
            break
    return dataFrame

class splitModule:
    def __init__(self, args) -> None:
        self.args = args
        self.genomesPath = args.genomesPath
        self.outputPath = args.outputPath
        self.name = args.name
        self.threads = args.threads

    def split_genomes(self):
        dataFrame = pd.read_csv(self.genomesPath)
        
        with Pool(self.threads) as p:
            dataFrames = list(tqdm(p.imap(split_genome, dataFrame.iterrows()), total=len(dataFrame), desc='Splitting', leave=True))
        splitDataFrame = pd.concat(dataFrames)

        # for index, row in tqdm(dataFrame.iterrows(), desc='Splitting', total=len(dataFrame), leave=True):
        #     genome = row['Genome']
        #     gene_seq = row['Gene_sequence']
        #     self.__split(genome, gene_seq, dataset)
        print("Saving to csv...")
        # splitDataFrame['sentence'] = splitDataFrame['sentence'].apply(lambda x: ' '.join(x))
        splitDataFrame['TDsentence'] = splitDataFrame['TDsentence'].apply(lambda x: ' '.join(x))
        splitDataFrame['TDlabels'] = splitDataFrame['TDlabels'].apply(lambda x:' '.join([str(i) for i in x]))
        # print(self.outputPath)
        os.makedirs(self.outputPath, exist_ok=True)
        splitDataFrame.to_csv(self.outputPath.joinpath(self.name+'_split.csv'), index=False)

# if __name__=='__main__':
#     parser = create_split_parser()
#     args = parser.parse_args()
#     split = splitModule(args)
#     split.split_genomes()




import argparse
import os
from pathlib import Path

from .baseCommand import baseCommand
from .extract import ExtractCommand, run
from .organize import organizeModule
from .split import splitModule
from .predict import genePredicter
from .output import outputFormatter
from .classify import geneClassifier

class piplineCommand(baseCommand):
    name = "pipeline"
    description = "Run the whole pipline of BGC-Prophet"

    def add_arguments(self, parser):
        parser.add_argument('--genomesDir', type=Path, required=True, help='genomes fasta directory')
        parser.add_argument('--lmdbPath', type=Path, required=False, default=None, help='lmdb path')
        parser.add_argument('--toks_per_batch', type=int, required=False, default=40960, help='toks per batch')
        parser.add_argument('--device', type=str, required=False, default='cpu', choices=['cpu', 'cuda'],help='device to use')
        parser.add_argument('--modelPath', type=Path, required=True, help='model path')
        parser.add_argument('--batch_size', type=int, required=False, default=512, help='batch size')
        parser.add_argument('--outputPath', type=Path, required=False, default=Path('./output/'), help='output path')
        parser.add_argument('--saveIntermediate', action='store_true', required=False, default=False, help='save intermediate results')
        parser.add_argument('--name', type=str, required=False, default='output', help='name of the output file')
        parser.add_argument('--threads', type=int, required=False, default=10, help='number of threads')
        parser.add_argument('--threshold', type=float, required=True, default=0.03, help='threshold of the prediction')
        parser.add_argument('--max_gap', type=int, required=True, default=3, help='max geme gap for the prediction')
        parser.add_argument('--min_count', type=int, required=True, default=2, help='min gene count for the prediction')
        parser.add_argument('--classifierPath', type=Path, required=True, help='classifier path')
        parser.add_argument('--classify_t', type=float, required=False, default=0.3, help='threshold for classification')

    def handle(self, args):
        if args.lmdbPath is None:
            parser_lmdb = argparse.ArgumentParser(prog="BGC-Prophet extract", description="Extract gene representations from genomes")
            ExtractCommand.add_arguments(ExtractCommand(), parser_lmdb)
            for file in os.listdir(args.genomesDir):
                if file.endswith('.fasta') or file.endswith('.faa'):
                    file_path = args.genomesDir.joinpath(file)
                    output_name = args.outputPath.joinpath('lmdb_' + args.name)
                    args_lmdb = parser_lmdb.parse_args('esm2_t6_8M_UR50D {file_path} {output_name} --toks_per_batch {toksNum} --include mean'.format(
                        file_path=file_path, output_name=output_name, toksNum=args.toks_per_batch).split())
                    run(args_lmdb)
            args.lmdbPath = str(args.outputPath.joinpath('lmdb_' + args.name))
        pipline = piplineModule(args)
        pipline.process()

class piplineModule:
    def __init__(self, args) -> None:
        self.args = args
        self.genomesDir = args.genomesDir
        self.lmdbPath = args.lmdbPath
        self.toks_per_batch = args.toks_per_batch
        self.device = args.device
        self.modelPath = args.modelPath
        self.batch_size = args.batch_size
        self.outputPath = args.outputPath
        self.saveIntermediate = args.saveIntermediate
        self.name = args.name

    def process(self, ):
        # organize genomes
        self.organize = organizeModule(self.args)
        self.organize.organize_genomes()
        self.organize.save()
        # split genomes
        self.args.genomesPath = self.outputPath.joinpath(self.name + '.csv')
        self.split = splitModule(self.args)
        self.split.split_genomes()
        # predict BGC
        self.args.datasetPath = self.outputPath.joinpath(self.name + '_split.csv')
        self.predict = genePredicter(self.args)
        self.predict.predict()
        self.predict.save()
        # output
        self.results = self.predict.results
        self.args.loadIntermediate = None
        self.output = outputFormatter(self.args)
        self.output.load()
        self.output.save(self.results)
        # classify BGC
        self.args.datasetPath = self.outputPath.joinpath(self.name + '.csv')
        self.classifier = geneClassifier(self.args)
        self.classifier.classify()
        self.classifier.process()
        self.classifier.save()
        # clear intermediate results
        if not self.saveIntermediate:
            os.remove(self.outputPath.joinpath(f'intermediate_{self.name}.npy'))
            os.remove(self.outputPath.joinpath(self.name + '_split.csv'))
            os.remove(self.outputPath.joinpath(self.name + '.csv'))
            os.remove(self.outputPath.joinpath(self.name + '_results.npy'))
        


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='BGC-Prophet process pipline.')
    # parser.add_argument('--genomesPath', type=str, required=True, help='genomes csv path')
    parser.add_argument('--genomesDir', type=str, required=True, help='genomes fasta directory')
    parser.add_argument('--lmdbPath', type=str, required=False, default=None, help='lmdb path')
    parser.add_argument('--toks_per_batch', type=int, required=False, default=40960, help='toks per batch')
    parser.add_argument('--device', type=str, required=False, default='cpu', help='device to use')
    parser.add_argument('--modelPath', type=str, required=True, help='model path')
    parser.add_argument('--batch_size', type=int, required=False, default=512, help='batch size')
    parser.add_argument('--outputPath', type=str, required=False, default='./output/', help='output path')
    parser.add_argument('--saveIntermediate', action='store_true', required=False, default=False, help='save intermediate results')
    parser.add_argument('--name', type=str, required=False, default='output', help='name of the output file')
    parser.add_argument('--threads', type=int, required=False, default=10, help='number of threads')
    parser.add_argument('--threshold', type=float, required=True, default=0.03, help='threshold of the prediction')
    parser.add_argument('--max_gap', type=int, required=True, default=3, help='max geme gap for the prediction')
    parser.add_argument('--min_count', type=int, required=True, default=2, help='min gene count for the prediction')
    parser.add_argument('--classifierPath', type=str, required=True, help='classifier path')
    parser.add_argument('--classify_t', type=float, required=False, default=0.3, help='threshold for classification')
    # parser.add_argument('--input', '-i', type=str, help='Input fasta file path.')

    args = parser.parse_args()
    if args.lmdbPath is None:
        parser_lmdb = argparse.ArgumentParser(prog="BGC-Prophet extract", description="Extract gene representations from genomes")
        ExtractCommand.add_arguments(ExtractCommand(), parser_lmdb)
        for file in os.listdir(args.genomesDir):
            if file.endswith('.fasta') or file.endswith('.faa'):
                args_lmdb = parser_lmdb.parse_args('esm2_t6_8M_UR50D {genomesDir}{filename} {outputDir}lmdb_{name} --toks_per_batch {toksNum} --include mean'.format(
                    genomesDir=args.genomesDir, filename=file, outputDir=args.outputPath, name=args.name, toksNum=args.toks_per_batch).split())
                run(args_lmdb)
        args.lmdbPath = args.outputPath + 'lmdb_' + args.name
    pipline = piplineModule(args)
    pipline.process()




from sklearn.metrics import roc_curve, auc, confusion_matrix
import numpy as np
import matplotlib.pyplot as plt

class final_validation:
    def __init__(self, x_true, x_pred, outpath) -> None:
        self.x_true = x_true
        self.x_pred = x_pred
        # self.y_true = y_true
        # self.y_pred = y_pred
        self.outpath = outpath

    def table(self, x_true, x_pred, label):
        # 将概率值转为预测标签
        x_pred = [1 if x >= 0.5 else 0 for x in x_pred]
        # 计算混淆矩阵
        cm = confusion_matrix(x_true, x_pred)
        print(f'confusion-matrix:\n{cm}')
        # 计算每个元素的百分比
        cm_percent = 100*cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]
        tn = cm_percent[0][0]
        fp = cm_percent[0][1]
        fn = cm_percent[1][0]
        tp = cm_percent[1][1]
        print('TP:\t{:.2f}%\tFN:\t{:.2f}%\nFP:\t{:.2f}%\tTN:\t{:.2f}%'.format(tp,fn,fp,tn))
        accuracy = 100*(tp+tn)/(tp+tn+fp+fn)
        recall = 100*tp/(tp+fn)
        precision = 100*tp/(tp+fp)
        F1_Score = 2*precision*recall/(precision+recall)
        print('Accuracy={:.2f}%\tRecall={:.2f}%\tprecision={:.2f}%\tF1-score={:2f}'.format(accuracy,recall,precision,F1_Score))

        # 绘制混淆矩阵
        fig, ax = plt.subplots(1,3,figsize=(20,8))
        # cm
        im0 = ax[0].imshow(cm, interpolation='nearest', cmap=plt.cm.Blues)
        fig.colorbar(im0, ax=ax[0])
        ticks = np.arange(2)
        ax[0].set_xticks(ticks)
        ax[0].set_yticks(ticks)
        ax[0].set_xticklabels(['0 (N)', '1 (P)'])
        ax[0].set_yticklabels(['0 (N)', '1 (P)'])
        ax[0].set_xlabel('Predicted label')
        ax[0].set_ylabel('True label')
        ax[0].set_title('Confusion Matrix',fontsize=16,fontweight ="bold")
        thresh = cm.max() / 2
        for i, j in np.ndindex(cm.shape):
            ax[0].text(j, i, format(cm[i, j]),
                     horizontalalignment="center",
                     color="white" if cm[i, j] > thresh else "black")
        # cm_percent
        im1 = ax[1].imshow(cm_percent, interpolation='nearest', cmap=plt.cm.Blues)
        fig.colorbar(im1, ax=ax[1])
        ticks = np.arange(2)
        ax[1].set_xticks(ticks)
        ax[1].set_yticks(ticks)
        ax[1].set_xticklabels(['0 (N)', '1 (P)'])
        ax[1].set_yticklabels(['0 (N)', '1 (P)'])
        ax[1].set_xlabel('Predicted label')
        ax[1].set_ylabel('True label')
        ax[1].set_title('Percent Confusion Matrix',fontsize=16,fontweight ="bold")
        thresh = cm_percent.max() / 2
        for i, j in np.ndindex(cm_percent.shape):
            ax[1].text(j, i, format(cm_percent[i, j], '.2f') + '%',
                     horizontalalignment="center",
                     color="white" if cm_percent[i, j] > thresh else "black")
        # table
        cell_text = [['{:.2f}%'.format(accuracy)],['{:.2f}%'.format(recall)],['{:.2f}%'.format(precision)],['{:.2f}%'.format(F1_Score)]]
        row_labels = ['Weighted Accuracy','Weighted Recall','Weighted Precision','Weighted F1-score']
        col_widths = [0.5, 0.5]
        tb = ax[2].table(cellText=cell_text, rowLabels=row_labels, loc='right', cellLoc='center', colWidths=col_widths, rowColours =["palegreen"]*4)
        tb.auto_set_column_width(True)
        tb.set_fontsize(14)  # 设置字体大小为14
        ax[2].axis('off')
        ax[2].set_title('Weighted Evaluation',fontsize=16,fontweight ="bold")
        # 保存图片
        file_path = self.outpath + f'_{label}_cm.png'
        fig.savefig(file_path, dpi=600, bbox_inches='tight')
    
    #def F1_score(self,x_true, x_pred):# 按照个数来计算
    #    x_pred = [int(item>=1) for  item in x_pred]
    #    f1 = f1_score(x_true.astype('int'),x_pred)
    #    print("F1-Score:{:.4f}".format(f1))

    def plot_step(self, x_true, x_pred, x):
        # x_true是真实标签，x_pred是模型输出的概率值
        # 计算TPR和FPR
        fpr, tpr, thresholds = roc_curve(x_true.astype('int'), x_pred)
        # 计算AUC
        roc_auc = auc(fpr, tpr)
        # 绘制ROC曲线
        plt.figure(figsize=(8, 8))
        plt.plot(fpr, tpr, color='blue', lw=2, label='ROC curve (area = %0.2f)' % roc_auc)
        plt.plot([0, 1], [0, 1], color='black', lw=2, linestyle='--')
        plt.xlim([-0.05, 1.05])
        plt.ylim([-0.05, 1.05])
        plt.xlabel('False Positive Rate')
        plt.ylabel('True Positive Rate')
        plt.title(f'{x} ROC')
        plt.legend(loc="lower right")
        file_path = self.outpath + f'_{x}_ROC.png'
        print("AUC:%0.3f"% roc_auc)
        plt.savefig(file_path, dpi=300)

    def plot_two(self):
        # x_true是真实标签，x_pred是模型输出的概率值
        # 计算TPR和FPR
        fpr = dict()
        tpr = dict()
        roc_auc = dict()
        fpr['gene'], tpr['gene'], _ = roc_curve(self.x_true.astype('int'), self.x_pred)
        fpr['gene'], tpr['gene'], _ = roc_curve(self.x_true.astype('int'), self.x_pred)
        # 计算AUC
        # roc_auc['bgc'] = auc(fpr['bgc'], tpr['bgc'])
        roc_auc['gene'] = auc(fpr['gene'], tpr['gene'])
        # 绘制ROC曲线
        plt.figure(figsize=(8, 8))
        # plt.plot(fpr['bgc'], tpr['bgc'], color='blue', lw=2, label=' BGC ROC curve (area = %0.2f)' % roc_auc['bgc'])
        plt.plot(fpr['gene'], tpr['gene'], color='red', lw=2, label=' Gene ROC curve (area = %0.2f)' % roc_auc['gene'])
        plt.plot([0, 1], [0, 1], color='black', lw=2, linestyle='--')
        plt.xlim([-0.05, 1.05])
        plt.ylim([-0.05, 1.05])
        plt.xlabel('False Positive Rate')
        plt.ylabel('True Positive Rate')
        plt.title('BGC and Gene ROC')
        plt.legend(loc="lower right")
        file_path = self.outpath + f'_Gene_ROC.png'
        plt.savefig(file_path, dpi=300)

    def result(self):
        # print('Final validation of BGC:')
        # self.plot_step(self.x_true,self.x_pred,'BGC')
        # self.table(self.x_true,self.x_pred,'BGC')
        print('Final validation of Gene:')
        self.plot_step(self.x_true,self.x_pred,'Gene')
        self.table(self.x_true,self.x_pred,'Gene')
        # self.plot_two()




import torch
import argparse
from torch.utils.data import DataLoader
import os
from tqdm import tqdm
import numpy as np
import pickle
import random

from torch.utils.tensorboard import SummaryWriter   
from data import DataReader, BGCLabelsDataset
from model import transformerEncoderNet
from loss import trainLoss
from utils import evaluate
from final_val import final_validation

class transformerEncoderTDeval:

    def __init__(self, args, writer, data, model, loss) -> None:
        self.args = args
        self.writer = writer
        self.data = data
        self.model = model
        self.loss = loss

        self.lmdbPath = args.lmdbPath
        self.batch_size = args.batch_size
        self.save_dir = args.save_dir
        self.transformerEncoderPath = args.transformerEncoderPath
        self.name = args.name
        self.results = []
        self.distribution = []
        
        # self.model = torch.load(self.lstmPath)
        self.model = model
        self.dataset = BGCLabelsDataset(data, lmdbPath=self.lmdbPath, mode='eval')
        self.dataLoader = DataLoader(dataset=self.dataset, batch_size=self.batch_size, shuffle=False, num_workers=5)

        self.device = 'cuda' if torch.cuda.is_available() else 'cpu'
        self.model.to(self.device)

        os.makedirs(self.save_dir, exist_ok=True)

    def eval(self):
        self.model.eval()
        total_TD_loss = 0
        total_TD_acc = 0
        with torch.no_grad():
            for i, data in tqdm(enumerate(self.dataLoader), desc='Evaluate', leave=True):
                sentence, labels, distribution = data[0], data[1], data[2]
                sentence = sentence.to(self.device)
                labels = labels.to(self.device)
                distribution = distribution.to(self.device)

                outputsTD = self.model(sentence) 
                # loss = self.loss(outputs, labels)
                # total_loss += loss.item()
                # labelloss = self.loss(outputsLabels, labels)
                TDLoss = self.loss(outputsTD, distribution)
                # total_label_loss += labelloss
                total_TD_loss += TDLoss

                # Label_correct = evaluate(outputsLabels.clone().detach(), labels)
                # Label_accuracy = Label_correct*100/torch.sum(labels).item()
                TD_correct = evaluate(outputsTD.clone().detach(), distribution)
                if torch.sum(distribution).item()!=0:
                    TD_accuracy = TD_correct*100/torch.sum(distribution).item()
                else:
                    predict = outputsTD.clone().detach()
                    predict[predict>0.5] = 1.
                    predict[predict<0.5] = 0.
                    TD_accuracy = torch.sum(torch.eq(predict, distribution)).item()/distribution.numel()


                # total_label_acc += Label_accuracy
                total_TD_acc += TD_accuracy
                if (i%10 == 0):
                    print(f'{i}/{len(self.dataLoader)} TD_accuracy: {TD_accuracy}')
                self.distribution.append(distribution.numpy(force=True))
                self.results.append(outputsTD.numpy(force=True)) 

            
        # self.total_label_acc = total_label_acc/len(self.dataLoader)
        self.total_TD_acc = total_TD_acc/len(self.dataLoader)
        self.distribution = np.vstack(self.distribution).flatten()
        self.results = np.vstack(self.results).flatten()
        print(f'distribution.shape: {self.distribution.shape}')
        print(f'result.shape: {self.results.shape}')

        # self.total_loss = total_loss/len(self.dataLoader)

        print('#TD_acc:%.3f' % self.total_TD_acc)

    def saveResults(self):
        # with open(self.save_dir + 'lstm_results.', 'w') as fp:
        np.save(self.save_dir + f'{self.name}_outputsTD.npy', self.results, allow_pickle=True)
        np.save(self.save_dir + f'{self.name}_distribution.npy', self.distribution, allow_pickle=True)
        self.writer.add_pr_curve('pr_curve', self.results, self.distribution, 0)
        self.plot = final_validation(x_true=self.distribution, x_pred=self.results, outpath=self.save_dir + self.name)
        self.plot.result()

            



if __name__=='__main__':
    parser = argparse.ArgumentParser(
        prog = 'transformerEncoder',
        description='transformerEncoder model to predict labels',
    )

    parser.add_argument('--lmdbPath', required=True)
    parser.add_argument('--datasetPath', required=True)

    parser.add_argument('--max_len', default=64, type=int)
    parser.add_argument('--left_padding', '-l', action='store_true', required=False)
    parser.add_argument('--nhead', type=int, required=True)
    parser.add_argument('--num_encoder_layers', required=True, default=4, type=int)
    parser.add_argument('--dropout', default=0.5, type=float)
    parser.add_argument('--batch_size', required=True, type=int)
    # parser.add_argument('--learning_rate', required=True, type=float)
    # parser.add_argument('--interval', required=False, default=10, type=int)
    # parser.add_argument('--epochs', required=True, type=int)
    parser.add_argument('--save_dir', default='./resultSave/')
    # parser.add_argument('--labels_num', default=8, type=int)
    parser.add_argument('--transformerEncoderPath', required=True)
    parser.add_argument('--name', required=True)
    # parser.add_argument()

    args = parser.parse_args()

    print('--------args----------')
    for k in list(vars(args).keys()):
        print('%s: %s' % (k, vars(args)[k]))
    print('--------args----------\n')

    writer = SummaryWriter('./log/TransformerEncoder/')
    data = DataReader(args.datasetPath, args.max_len)
    embedding_dim = data.embedding_dim
    # model = LSTMTimeDistributedNet(embedding_dim, args.hidden_dim, num_layers=args.num_layers, max_len=args.max_len, labels_num=args.labels_num, dropout=args.dropout)
    model = torch.load(args.transformerEncoderPath)
    loss = trainLoss()

    evaluater = transformerEncoderTDeval(args=args, writer=writer, data=data, model=model, loss=loss)
    evaluater.eval()
    evaluater.saveResults()



import torch
import torch.nn as nn

class trainLoss(nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.criterion = nn.BCELoss(reduction='mean')
        # self.criterion = nn.CrossEntropyLoss(label_smoothing=0.05)

    def forward(self, output, labels):
        # output: (batch_size, labels_num)
        return self.criterion(output, labels)



import torch
import math
from torch import nn
from torch import Tensor
import torch.nn.functional as F
from typing import Optional, Any, Union, Callable

from .model import PositionalEncoding

class transformerClassifier(nn.Module):
    def __init__(self, d_model: int=512, nhead: int=8, num_encoder_layers: int=6, max_len: int=128,
                 dim_feedforward: int=2048, labels_num: int=7, transformer_dropout: float=0.1, mlp_dropout: float=0.5,
                 activation: Union[str, Callable[[Tensor], Tensor]]=F.gelu,
                 layer_norm_eps: float=1e-5, batch_first: bool=True, norm_first: bool=True,
                 device=None, dtype=None) -> None:
        super().__init__()
        factory_kwargs = {'device': device, 'dtype': dtype}
        self.pos_encoder = PositionalEncoding(num_hiddens=d_model)
        encoder_layer = nn.TransformerEncoderLayer(d_model, nhead, dim_feedforward, 
                                                   transformer_dropout, activation, layer_norm_eps, 
                                                   batch_first, norm_first, **factory_kwargs)
        encoder_norm = nn.LayerNorm(d_model, eps=layer_norm_eps, **factory_kwargs)
        self.encoder = nn.TransformerEncoder(encoder_layer, num_encoder_layers, encoder_norm)

        self.classifier = nn.Sequential(
            nn.Dropout(mlp_dropout),
            # nn.Linear(d_model, 128),
            # nn.GELU(),
            # nn.Linear(128, 32),
            # nn.GELU(),
            # nn.Linear(32, labels_num),
            nn.Linear(d_model, labels_num),
            nn.Sigmoid()
        )

        self.__reset_parameters()
        self.d_model = d_model
        self.nhead = nhead
        self.batch_first = batch_first

    def forward(self, src: Tensor, src_key_padding_mask: Tensor) -> Any:
        # src: (batch_size, max_len, embed_dim)
        memory = self.pos_encoder(src)
        memory = self.encoder(src, mask=None, src_key_padding_mask=src_key_padding_mask)
        # memory = memory*src_key_padding_mask.unsqueeze(-1)
        # memory: (batch_size, max_len, embed_dim), src_key_padding_mask: (batch_size, max_len, embed_dim)
        # memory = memory[:, -1, :].squeeze(1)
        memory = memory.mean(dim=1).squeeze(1)
        labels = self.classifier(memory)
        return labels

    def __reset_parameters(self):
        r"""Initiate parameters in the transformer model."""
        for p in self.parameters():
            if p.dim() > 1:
                nn.init.xavier_uniform_(p)




import torch
import torch.nn as nn
import torch.nn.functional as F

class FocalLoss(nn.Module):
    def __init__(self, alpha, gamma, reduction='mean'):
        super().__init__()
        self.alpha = alpha
        self.gamma = gamma
        self.reduction = reduction

    def forward(self, inputs, targets):
        inputs = inputs.view(-1)
        targets = targets.view(-1)

        BCE = F.binary_cross_entropy(inputs, targets, reduction=self.reduction)
        BEC_EXP = torch.exp(-BCE)
        focal_loss = self.alpha * (1-BEC_EXP) ** self.gamma * BCE

        return focal_loss




import torch
from torch.utils.data import DataLoader
from loss import *
from data import DataReader, BGCLabelsDataset
import argparse
import os
from tqdm import tqdm
import numpy as np
from utils import evaluate
from sklearn.metrics import roc_curve, auc, confusion_matrix

class evalAll:
    def __init__(self, args, data) -> None:
        self.args = args
        self.data = data
        self.models_folder = args.models_folder
        # self.save_dir = args.save_dir
        self.epochs = args.epochs
        self.lmdbPath = args.lmdbPath
        self.batch_size = args.batch_size
        # self.models = os.listdir(self.models)
        self.device = 'cuda' if torch.cuda.is_available() else 'cpu'
        # os.makedirs(self.save_dir, exist_ok=True)

    def eval_all(self):
        loss = trainLoss()
        self.dataset = BGCLabelsDataset(self.data, lmdbPath=self.lmdbPath, mode='eval')
        for i in range(self.epochs):
            modelPath = self.models_folder+f'transformerEncoder_Model_TD_{i}.pt'
            model = torch.load(modelPath)
            print(f"#Epoch: {i}")
            self.eval_one(model=model, loss=loss)    

    def eval_one(self, model, loss):
        model.to(self.device)
        self.dataLoader = DataLoader(dataset=self.dataset, batch_size=self.batch_size, shuffle=False, num_workers=5)
        model.eval()
        total_TD_loss = 0
        distributions = []
        results = []
        total_TD_acc = 0
        with torch.no_grad():
            # print(len(self.dataLoader))
            # print(type(self.dataLoader))
            for i, data in tqdm(enumerate(self.dataLoader), desc='Evaluate', leave=True):
                sentence, labels, distribution = data[0], data[1], data[2]
                sentence = sentence.to(self.device)
                labels = labels.to(self.device)
                distribution = distribution.to(self.device)

                outputsTD = model(sentence) 
                TDLoss = loss(outputsTD, distribution)
                total_TD_loss += TDLoss

                # Label_correct = evaluate(outputsLabels.clone().detach(), labels)
                # Label_accuracy = Label_correct*100/torch.sum(labels).item()
                TD_correct = evaluate(outputsTD.clone().detach(), distribution)
                if torch.sum(distribution).item()!=0:
                    TD_accuracy = TD_correct*100/torch.sum(distribution).item()
                else:
                    predict = outputsTD.clone().detach()
                    predict[predict>0.5] = 1.
                    predict[predict<0.5] = 0.
                    TD_accuracy = torch.sum(torch.eq(predict, distribution)).item()/distribution.numel()


                # total_label_acc += Label_accuracy
                total_TD_acc += TD_accuracy
                if (i%10 == 0):
                    print(f'{i}/{len(self.dataLoader)} TD_accuracy: {TD_accuracy}')
                distributions.append(distribution.numpy(force=True))
                results.append(outputsTD.numpy(force=True)) 

            
        # self.total_label_acc = total_label_acc/len(self.dataLoader)
        total_TD_acc = total_TD_acc/len(self.dataLoader)
        distributions = np.vstack(distributions).flatten()
        results = np.vstack(results).flatten()
        print(f'distribution.shape: {distributions.shape}')
        print(f'result.shape: {results.shape}')

        # self.total_loss = total_loss/len(self.dataLoader)

        print('#TD_acc:%.3f' % total_TD_acc)
        fpr, tpr, _ = roc_curve(distributions.astype('int'), results)
        roc_auc = auc(fpr, tpr)
        print(f"AUC: {roc_auc}")
        predicts = results.copy()
        predicts[predicts>=0.5] = 1.
        predicts[predicts<0.5] = 0.
        # 计算混淆矩阵
        cm = confusion_matrix(distributions, predicts)
        print(f'confusion-matrix:\n{cm}')
        # 计算每个元素的百分比
        cm_percent = 100*cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]
        tn = cm_percent[0][0]
        fp = cm_percent[0][1]
        fn = cm_percent[1][0]
        tp = cm_percent[1][1]
        print('TP:\t{:.2f}%\tFN:\t{:.2f}%\nFP:\t{:.2f}%\tTN:\t{:.2f}%'.format(tp,fn,fp,tn))
        accuracy = 100*(tp+tn)/(tp+tn+fp+fn)
        recall = 100*tp/(tp+fn)
        precision = 100*tp/(tp+fp)
        F1_Score = 2*precision*recall/(precision+recall)
        print('Accuracy={:.2f}%\tRecall={:.2f}%\tprecision={:.2f}%\tF1-score={:2f}'.format(accuracy,recall,precision,F1_Score))


            


if __name__=='__main__':
    parser = argparse.ArgumentParser(
        prog = 'Evaluate_all',
        description='evaluate all trained transformer encoder model',
    )
    parser.add_argument('--models_folder', required=True, type=str)
    # parser.add_argument('--save_dir', required=True, type=str)
    parser.add_argument('--epochs', required=True, type=int)
    parser.add_argument('--lmdbPath', required=True, type=str)
    parser.add_argument('--datasetPath', required=True, type=str)
    parser.add_argument('--max_len', required=True, type=int)
    parser.add_argument('--batch_size', required=True, type=int)

    args = parser.parse_args()

    print('--------args----------')
    for k in list(vars(args).keys()):
        print('%s: %s' % (k, vars(args)[k]))
    print('--------args----------\n')

    data = DataReader(args.datasetPath, args.max_len)


    evaluater = evalAll(args, data=data)
    evaluater.eval_all()




import math
from typing import Any

import torch 
import torch.nn as nn
from torch import Tensor
from torch.nn.parameter import Parameter, UninitializedParameter

class batch3Linear(nn.Module):
    r'''
    batch linear layer
    '''
    __constants__ = ['batch_size', 'in_features', 'out_features']
    batch_size:int
    in_features:int
    out_features:int
    weight:Tensor

    def __init__(self, batch_size:int, in_features:int, out_features:int, bias:bool=True,
                 device=None, dtype=None) -> None:
        factory_kwargs = {'device':device, 'dtype':dtype}
        super().__init__()
        self.batch_size = batch_size
        self.in_features = in_features
        self.out_features = out_features
        self.weight = Parameter(torch.empty((batch_size, in_features, out_features), **factory_kwargs))
        if bias:
            self.bias = Parameter(torch.empty((batch_size, 1, out_features), **factory_kwargs))
        else:
            self.register_parameter('bias', None)
        self.reset_parameters()

    def reset_parameters(self) -> None:
        # Setting a=sqrt(5) in kaiming_uniform is the same as initializing with
        # uniform(-1/sqrt(in_features), 1/sqrt(in_features)). For details, see
        # https://github.com/pytorch/pytorch/issues/57109
        nn.init.kaiming_uniform_(self.weight, a=math.sqrt(5))
        if self.bias is not None:
            if self.bias is not None:
                fan_in, _ = nn.init._calculate_fan_in_and_fan_out(self.weight)
                bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
                nn.init.uniform_(self.bias, -bound, bound)

    def forward(self, input:Tensor) -> Tensor:
        assert input.shape[0]==self.weight.shape[0], 'batch_size not equal!'
        # input: (max_len, batch_size in_features)
        # self.weight: (max_len, in_features, out_features)
        # return: (max_len, batch_size, out_features)
        return torch.bmm(input, self.weight) + self.bias
    
    def extra_repr(self) -> str:
        return 'batch_size={}, in_features={}, out_features={}, bias={}'.format(
            self.batch_size, self.in_features, self.out_features, self.bias is not None
        )





class baseCommand:
    """
    Base command class for BGC-Prophet
    """
    def add_arguments(self, parser):
        """
        Add arguments to the parser
        """
        raise NotImplementedError

    def handle(self, args):
        """
        Handle the command
        """
        raise NotImplementedError
    


import torch
import math
from torch import nn
from torch import Tensor
import torch.nn.functional as F
from typing import Optional, Any, Union, Callable

from batch3Linear import batch3Linear
# from torch.nn.module import Transformer


class LSTMNet(nn.Module):
    def __init__(self, embedding_dim:int,  
                 hidden_dim:int, num_layers:int, 
                 dropout:int=0.1, initWeight:bool=False) -> None:
        super().__init__()

        self.LSTM = nn.LSTM(embedding_dim, hidden_dim, num_layers=num_layers, batch_first=True, bidirectional=True)
        # 全连接层的权重相互独立，用于预测每个基因是否是BGC的组成部分
        self.TD = nn.Sequential(
            nn.LayerNorm(hidden_dim*2),
            nn.Dropout(dropout),
            nn.Linear(in_features=hidden_dim*2, out_features=128),
            nn.GELU(),
            nn.Linear(in_features=128, out_features=32),
            nn.GELU(),
            nn.Linear(in_features=32, out_features=1),
            nn.Sigmoid()
        )
        # self.classifier = nn.Sequential(
        #     nn.LayerNorm(hidden_dim*2),
        #     nn.Dropout(dropout),
        #     nn.Linear(in_features=hidden_dim*2, out_features=64), 
        #     nn.GELU(),
        #     nn.Dropout(dropout),
        #     nn.Linear(in_features=64, out_features=labels_num),
        #     nn.Sigmoid()
        # )
        if initWeight:
            self.__initWeight()

    def forward(self, inputs):
        # x = self.MAblock(inputs)
        # x = self.attention(inputs)
        # print(x.shape)
        x, _ = self.LSTM(inputs, None)
        TD = self.TD(x)
        # memory = torch.sum(x, dim=1) / x.shape[1]
        # labels = self.classifier(memory)
        # 输出是否是BGC的组成部分，以及BGC种类两个信息
        return TD.squeeze()
    
    def __initWeight(self):
        pass



    
    


import numpy as np
import pandas as pd
import argparse
from pathlib import Path

from .baseCommand import baseCommand

class outputCommand(baseCommand):
    name = "output"
    description = "Output the results of the prediction"

    def add_arguments(self, parser):
        parser.add_argument('--datasetPath', type=Path, required=True, help='dataset path')
        parser.add_argument('--outputPath', type=Path, required=False, default=Path('./output/'), help='output path')
        parser.add_argument('--loadIntermediate', type=str, required=True, default=None, help='load intermediate results')
        parser.add_argument('--name', type=str, required=False, default='output', help='name of the output file')
        parser.add_argument('--threshold', type=float, required=True, default=0.5, help='threshold for the prediction')
        parser.add_argument('--max_gap', type=int, required=True, default=3, help='max gene gap for the prediction')
        parser.add_argument('--min_count', type=int, required=True, default=2, help='min gene count for the prediction')
    
    def handle(self, args):
        output = outputFormatter(args)
        output.load()
        output.save(output.results)


class outputFormatter:
    def __init__(self, args) -> None:
        self.datasetPath = args.datasetPath
        self.outputPath = args.outputPath
        self.loadIntermediate = args.loadIntermediate
        self.name = args.name
        self.threshold = args.threshold
        self.max_gap = args.max_gap
        self.min_count = args.min_count
        # self.dataFrame = pd.DataFrame(columns=['ID', 'sentence', 'labels', 'isBGC', 'TDsentence', 'TDlabels'])

    def load(self):
        self.dataset = pd.read_csv(self.datasetPath)
        if self.loadIntermediate:
            self.results = np.load(self.loadIntermediate)

    def organize(self, results):
        self.dataset['result'] = results.tolist()
        # print(self.dataset.head())
        outputDataset = self.dataset.apply(self.output, axis=1, result_type='expand')
        self.dataFrame = self.dataset.loc[:, ['ID', 'TDsentence', 'labels', 'isBGC']]
        # print(self.dataFrame.head())
        self.dataFrame['sentence'] = outputDataset['sentence'].copy()
        self.dataFrame['TDlabels'] = outputDataset['TDlabels'].copy()
        self.dataFrame['isBGC'] = outputDataset['isBGC'].copy()
        # print(self.dataFrame.head())
        self.dataFrame['sentence'] = self.dataFrame['sentence'].apply(lambda x: ' '.join(x) if x!=None else None)
        self.dataFrame['TDlabels'] = self.dataFrame['TDlabels'].apply(lambda x: ' '.join([str(i) for i in x]))
        # print(self.dataFrame.head())

    def output(self, series):
        TDlabels = []
        result = series['result']
        TDsentence = series['TDsentence'].split()
        result = np.array(result)
        result[result>self.threshold] = 1.
        result[result<=self.threshold] = 0.
        TDlabels = result.tolist()
        if not 1. in TDlabels:
            TDlabels = [0]*len(TDlabels)
            return pd.Series([None, 'No', TDlabels], index=['sentence', 'isBGC', 'TDlabels'])
        current_start = TDlabels.index(1.)
        current_end = current_start
        output_start = 0
        output_end = 0
        max_range = 0
        for i in range(current_start+1, len(TDlabels)):
            if TDlabels[i]==1:
                if i-current_end<=self.max_gap:
                    current_end = i
                else:
                    if current_end-current_start>max_range:
                        output_start = current_start
                        output_end = current_end
                        max_range = current_end-current_start
                    current_start = i
                    current_end = i
        if current_end-current_start>max_range:
            output_start = current_start
            output_end = current_end
            max_range = current_end-current_start
        if max_range<self.min_count:
            isBGC = 'No'
            sentence = None
            TDlabels = [0]*len(TDlabels)
        else:
            sentence = TDsentence[output_start:output_end+1]
            isBGC = 'Yes'
            TDlabels = [0]*output_start+[1]*(output_end-output_start+1)+[0]*(len(TDlabels)-output_end-1)            
        return pd.Series([sentence, isBGC, TDlabels], index=['sentence', 'isBGC', 'TDlabels'])

    def save(self, results):
        self.organize(results)
        self.dataFrame.to_csv(self.outputPath.joinpath(f'{self.name}.csv'), index=False, header=True)

# if __name__=='__main__':
#     parser = argparse.ArgumentParser(
#         prog='output',
#         description='Output the results of the prediction',
#     )
#     parser.add_argument('--datasetPath', type=str, required=True, help='dataset path')
#     parser.add_argument('--outputPath', type=str, required=False, default='./output/', help='output path')
#     parser.add_argument('--loadIntermediate', type=str, required=True, default=None, help='load intermediate results')
#     parser.add_argument('--name', type=str, required=False, default='output', help='name of the output file')
#     parser.add_argument('--threshold', type=float, required=True, default=0.5, help='threshold for the prediction')
#     parser.add_argument('--max_gap', type=int, required=True, default=3, help='max gene gap for the prediction')
#     parser.add_argument('--min_count', type=int, required=True, default=2, help='min gene count for the prediction')

#     args = parser.parse_args()

#     output = outputFormatter(args)
#     # output = outputFormatter(args.datasetPath, args.outputPath, args.loadIntermediate, args.name, args.threshold, args.max_gap, args.min_count)
#     output.load()
#     output.save(output.results)



import argparse
import pandas as pd
import os
from multiprocessing import Pool
from tqdm import tqdm
from functools import partial
from .baseCommand import baseCommand
from pathlib import Path
from Bio import SeqIO


def organize_genome(genome, genomesDir):
    genome_name = os.path.splitext(genome)[0]
    # gene_seq = []
    genome_seq = list(SeqIO.parse(genomesDir.joinpath(genome), 'fasta'))
    gene_names = [seq.id for seq in genome_seq]
    # with open(genomesDir.joinpath(genome), 'r') as f:
    #     for line in f:
    #         if not line.startswith('>'):
    #             continue
    #         else:
    #             gene_seq.append(line.strip()[1:])
    return pd.Series([genome_name, gene_names], index=['Genome', 'Gene_sequence'])

class organizeCommand(baseCommand):
    name = 'organize'
    description = 'Organize the input genomes of BGC-Prophet'

    def add_arguments(self, parser):
        parser.add_argument('--genomesDir', type=Path, required=True, help='genomes fasta directory')
        parser.add_argument('--outputPath', type=Path, required=False, default=Path('./output/'), help='output path')
        parser.add_argument('--name', type=str, required=False, default='output', help='name of the output file')
        parser.add_argument('--threads', type=int, required=False, default=10, help='number of cpu threads to use in organizing')

    def handle(self, args):
        organize = organizeModule(args)
        organize.organize_genomes()
        organize.save()

class organizeModule:
    def __init__(self, args) -> None:
        self.args = args
        self.genomesDir = args.genomesDir
        self.outputPath = args.outputPath
        self.name = args.name
        self.threads = args.threads
        
    def organize_genomes(self, ):
        self.dataFrame = pd.DataFrame(columns=['Genome', 'Gene_sequence'])
        files = os.listdir(self.genomesDir)
        files = [file for file in files if file.endswith('.fasta') or file.endswith('.faa')]
        organize_genome_Dir = partial(organize_genome, genomesDir=self.genomesDir)
        with Pool(self.threads) as p:
            dataFrames = list(tqdm(p.imap(organize_genome_Dir, files), total=len(files), desc='Organizing', leave=True))
        self.dataFrame = pd.DataFrame(dataFrames)

    def save(self, ):
        print("Saving to csv...")
        os.makedirs(self.outputPath, exist_ok=True)
        self.dataFrame.to_csv(self.outputPath.joinpath(self.name+'.csv'), index=False)


# if __name__=='__main__':
#     parser = create_organize_parser()
#     args = parser.parse_args()
#     organize = organizeModule(args)
#     organize.organize_genomes()
#     organize.save()



from .classifier import transformerClassifier
from .data import DataReader, BGCLabelsDataset
from .final_val import final_validation
from .focal_loss import FocalLoss
from .loss import trainLoss
from .model import transformerEncoderNet




from .classify import classifyCommand
from .extract import ExtractCommand
from .organize import organizeCommand
from .output import outputCommand
from .pipline import piplineCommand
from .predict import predictCommand
from .split import splitCommand

__all__ = ['classifyCommand', 'ExtractCommand', 'organizeCommand', 'outputCommand', 'piplineCommand', 'predictCommand', 'splitCommand']


import torch
import argparse
import torch.optim as optim
from torch.utils.data import DataLoader
from torch.utils.tensorboard import SummaryWriter   
from tqdm import tqdm
import os
import pickle
import numpy as np
import random

from data import DataReader, BGCLabelsDataset
from loss import trainLoss
from utils import evaluate
# from model import transformerEncoderNet
# from classifyTrainer import TransformerClassifier
from classifier import transformerClassifier
from sklearn.metrics import roc_curve, auc, confusion_matrix

class classifyEval:
    def __init__(self, args, writer, data, model, loss) -> None:
        self.args = args
        self.writer = writer
        self.data = data
        self.model = model
        self.loss = loss

        self.lmdbPath = args.lmdbPath
        self.batch_size = args.batch_size
        self.save_dir = args.save_dir
        self.modelPath = args.modelPath
        self.name = args.name
        self.results = []
        self.labels = []

        self.dataset = BGCLabelsDataset(self.data, self.lmdbPath, 'eval')
        self.dataLoader = DataLoader(dataset=self.dataset, batch_size=self.batch_size, shuffle=False, num_workers=5)
        
        self.device = 'cuda' if torch.cuda.is_available() else 'cpu'
        self.model.to(self.device)

        os.makedirs(self.save_dir, exist_ok=True)

    def eval(self):
        self.model.eval()
        total_loss = 0
        total_acc = 0
        with torch.no_grad():
            for i, data in tqdm(enumerate(self.dataLoader), desc='Evaluate', leave=True):
                sentence, labels, distribution = data[0], data[1], data[2]
                sentence = sentence.to(self.device)
                labels = labels.to(self.device)
                distribution = distribution.to(self.device)

                outputs = self.model(sentence, distribution)
                lloss = self.loss(outputs, labels)
                total_loss += lloss.item()
                correct = evaluate(outputs.clone().detach(), labels)
                predict = outputs.clone().detach()
                predict[predict>=0.5] = 1
                predict[predict<0.5] = 0
                acc = torch.sum(torch.eq(predict, labels)).item() / labels.numel()
                total_acc += acc

                if (i%10==0):
                    print(f'{i}/{len(self.dataLoader)}: loss: {lloss.item()}, acc: {acc}')
                self.results.append(outputs.numpy(force=True))
                self.labels.append(labels.numpy(force=True))
        self.total_loss = total_loss / len(self.dataLoader)
        self.total_acc = total_acc / len(self.dataLoader)
        print(f'loss: {self.total_loss}, acc: {self.total_acc}')
        self.results = np.vstack(self.results)
        self.labels = np.vstack(self.labels)
        fpr, tpr, thresholds = roc_curve(self.labels.flatten().astype('int'), self.results.flatten())
        roc_auc = auc(fpr, tpr)
        print(f'roc_auc: {roc_auc}')
        predicts = self.results.copy()
        predicts[predicts>=0.5] = 1
        predicts[predicts<0.5] = 0
        tn, fp, fn, tp = confusion_matrix(self.labels.flatten().astype('int'), predicts.flatten()).ravel()
        print(f'tn: {tn}, fp: {fp}, fn: {fn}, tp: {tp}')
        accuracy = (tp+tn)/(tp+tn+fp+fn)
        recall = tp/(tp+fn)
        precision = tp/(tp+fp)
        f1 = 2*precision*recall/(precision+recall)
        print(f'accuracy: {accuracy}, recall: {recall}, precision: {precision}, f1: {f1}')

    def saveResults(self):
        np.save(self.save_dir + self.name + '_results.npy', self.results)
        np.save(self.save_dir + self.name + '_labels.npy', self.labels)

if __name__=='__main__':
    parser = argparse.ArgumentParser(
        prog='classifyEval',
        description='Evaluate the classification model',
    )
    parser.add_argument('--datasetPath', type=str, required=True, help='Path to the dataset')
    parser.add_argument('--lmdbPath', type=str, required=True, help='Path to the lmdb')
    parser.add_argument('--modelPath', type=str, required=True, help='Path to the model')
    parser.add_argument('--max_len', type=int, default=128, required=True)
    parser.add_argument('--save_dir', type=str, default='./resultSave/', required=False)
    parser.add_argument('--name', type=str, default='classifyEval', required=False)
    parser.add_argument('--batch_size', type=int, default=64, required=False)

    args = parser.parse_args()
    print('--------args----------')
    for k in list(vars(args).keys()):
        print('%s: %s' % (k, vars(args)[k]))
    print('--------args----------\n')
    writer = SummaryWriter('./log/classifyEval/')
    data = DataReader(args.datasetPath, args.max_len)
    embedding_dim = data.embedding_dim
    model = torch.load(args.modelPath)
    loss = trainLoss()
    
    evaluator = classifyEval(args=args, writer=writer, data=data, model=model, loss=loss)
    evaluator.eval()
    evaluator.saveResults()

    


import torch
import math
from torch import nn
from torch import Tensor
import torch.nn.functional as F
from typing import Optional, Any, Union, Callable

from .batch3Linear import batch3Linear
# from torch.nn.module import Transformer


class transformerEncoderNet(nn.Module):
    def __init__(self, d_model: int=512, nhead: int=8, num_encoder_layers: int=6, max_len: int=128, 
                 dim_feedforward: int=2048, transformer_dropout: float=0.1, mlp_dropout: float=0.5, 
                 activation: Union[str, Callable[[Tensor], Tensor]]=F.gelu,
                #  custom_encoder: Optional[Any]=None,
                 layer_norm_eps: float=1e-05, batch_first: bool=True, norm_first: bool=True,
                 device=None, dtype=None) -> None:
        super().__init__()
        factory_kwargs = {'device':device, 'dtype':dtype}
        self.pos_encoder = PositionalEncoding(num_hiddens=d_model)
        encoder_layer = nn.TransformerEncoderLayer(d_model, nhead, dim_feedforward, 
                                                   transformer_dropout, activation, layer_norm_eps, batch_first, norm_first,
                                                   **factory_kwargs)
        encoder_norm = nn.LayerNorm(d_model, eps=layer_norm_eps, **factory_kwargs)
        self.encoder = nn.TransformerEncoder(encoder_layer, num_encoder_layers, encoder_norm)

        # self.TDOut = FeatLayer(max_len, d_model, 1)
        self.out = nn.Sequential(
            nn.Dropout(mlp_dropout),
            nn.Linear(in_features=d_model, out_features=128),
            nn.GELU(),
            nn.Linear(in_features=128, out_features=32),
            nn.GELU(),
            nn.Linear(in_features=32, out_features=1),
            nn.Sigmoid()
        )


        self._reset_parameters()
        self.d_model = d_model
        self.nhead = nhead
        self.batch_first = batch_first

    def forward(self, src: Tensor):
        # src: (batch_size, max_len, embed_dim)
        # src = src * math.sqrt(self.d_model)
        src = self.pos_encoder(src)
        memory = self.encoder(src, mask=None, src_key_padding_mask=None)

        # memory_reshaped = memory.reshape(memory.shape[1], memory.shape[0], -1)
        # TD = self.TDOut(memory_reshaped)
        # TD = TD.reshape(memory.shape[0], memory.shape[1], -1).squeeze(-1)
        TD = self.out(memory).squeeze(-1)
        return TD
    
    def _reset_parameters(self):
        r"""Initiate parameters in the transformer model."""
        for p in self.parameters():
            if p.dim() > 1:
                nn.init.xavier_uniform_(p)



class FeatLayer(nn.Module):
    def __init__(self, max_len: int, dim_in: int, dim_out: int) -> None:
        super().__init__()
        sizes = []
        while dim_in>dim_out:
            sizes.append(dim_in)
            dim_in //=4
        self.feat = nn.Sequential(
            nn.LayerNorm(sizes[0]),
            nn.Dropout(0.2),
            *[nn.Sequential(batch3Linear(batch_size=max_len, in_features=sizes[i], 
                                         out_features=sizes[i+1]), nn.GELU()) for i in range(len(sizes)-1)],
            batch3Linear(batch_size=max_len, in_features=sizes[-1], out_features=dim_out),
            nn.Sigmoid()                                
        )
        
    def forward(self, inputs):
        return self.feat(inputs)


class PositionalEncoding(nn.Module):
    """位置编码"""
    def __init__(self, num_hiddens:int, dropout:float=0.1, max_len:int=1000):
        super().__init__()
        self.dropout = nn.Dropout(dropout)
        # 创建一个足够长的P
        self.P = torch.zeros((1, max_len, num_hiddens))
        X = torch.arange(max_len, dtype=torch.float32).reshape(
            -1, 1) / torch.pow(10000, torch.arange(
            0, num_hiddens, 2, dtype=torch.float32) / num_hiddens)
        self.P[:, :, 0::2] = torch.sin(X)
        self.P[:, :, 1::2] = torch.cos(X)

    def forward(self, X):
        X = X + self.P[:, :X.shape[1], :].to(X.device)
        return self.dropout(X)


import torch
import argparse
import torch.optim as optim
from torch.utils.data import DataLoader
from torch.utils.tensorboard import SummaryWriter   
from tqdm import tqdm
import os
import pickle
import numpy as np
import random
from timm.scheduler.cosine_lr import CosineLRScheduler

from data import DataReader, BGCLabelsDataset
from loss import trainLoss
from utils import evaluate
from model import transformerEncoderNet
# from lstm_model import LSTMNet



# 预测每个基因是否属于BGC
class TransformerEncoderTrainer:
    def __init__(self, args, writer, data, model, TimeDistributedLoss) -> None:
        self.args = args
        self.writer = writer
        self.d_model = data.embedding_dim
        self.lmdbPath = args.lmdbPath
        self.batch_size = args.batch_size
        self.interval = args.interval
        self.ditribute_epochs = args.distribute_epochs
        self.warmup_epochs = args.warmup_epochs
        self.save_dir = args.save_dir
        self.learning_rate = args.learning_rate
        # self.savePath = args.save_dir

        self.data = data
        self.train_dataset = BGCLabelsDataset(self.data, self.lmdbPath, mode='train')
        print('Train Data length: ', len(self.train_dataset))
        self.test_dataset = BGCLabelsDataset(self.data, self.lmdbPath, mode='test')
        print('Train Data length: ', len(self.test_dataset))
        self.train_dataLoader = DataLoader(dataset=self.train_dataset, batch_size=self.batch_size, shuffle=True, num_workers=5)
        self.test_dataLoader = DataLoader(dataset=self.test_dataset, batch_size=self.batch_size, shuffle=True, num_workers=5)
        self.model = model
        self.TimeDistributedLoss = TimeDistributedLoss
        print(torch.cuda.is_available())
        self.device = 'cuda' if torch.cuda.is_available() else 'cpu'
        self.model.to(self.device)
        
        self.save_path = self.save_dir + \
        f'transformerEncoder_TD_loss/bS_{self.batch_size}_dE_{self.ditribute_epochs}_lR_{self.learning_rate}_mL_{args.max_len}_d_{self.d_model}_nH_{args.nhead}_nEL_{args.num_encoder_layers}_tdP_{args.transformer_dropout}_mdP_{args.mlp_dropout}_TD/'
        os.makedirs(self.save_path, exist_ok=True)
        # with open(self.save_path+'labels_list.pkl', 'wb') as fp:
        #     pickle.dump(self.data.labels_list, fp)
        #     print("Save labels_list")
        print(self.model)
        # self.writer.add_graph(model=self.model, input_to_model=torch.randn(self.batch_size, args.max_len, data.embedding_dim))

    def train_TD_step(self, epoch):
        self.model.train()
        total_TD_loss = 0
        total_TD_acc = 0
        for i, data in tqdm(enumerate(self.train_dataLoader), desc='Train TD', leave=True):
            sentence, labels, distribution = data[0], data[1], data[2]
            sentence = sentence.to(self.device)
            labels = labels.to(self.device)
            distribution = distribution.to(self.device)

            self.optimizer.zero_grad()
            outputsTD = self.model(sentence)
            TDLoss = self.TimeDistributedLoss(outputsTD, distribution)

            total_TD_loss += TDLoss

            # 计算准确度
            TD_correct = evaluate(outputsTD.clone().detach(), distribution)
            if torch.sum(distribution).item() <= 0.0:
                TD_accuracy = 100.0
            else:
                TD_accuracy = TD_correct*100/torch.sum(distribution).item()

            # Label_accuracy = Label_correct*100/labels.numel()
            # TD_accuracy = TD_correct*100/distribution.numel()

            total_TD_acc += TD_accuracy

            TDLoss.backward()
            self.optimizer.step()

            if (i%self.interval == 0):
                print('#Epoch:%d, %d/%d Loss:%.5f dacc:%.3f' % (epoch, i, len(self.train_dataLoader), 
                    total_TD_loss/self.interval, total_TD_acc/self.interval))
                self.train_total_TD_loss += total_TD_loss
                self.train_total_TD_acc += total_TD_acc
                total_TD_loss = 0
                total_TD_acc = 0
        self.train_total_TD_acc /= len(self.train_dataLoader)
        self.train_total_TD_loss /= len(self.train_dataLoader)
        torch.save(self.model, self.save_path + f'transformerEncoder_Model_TD_{epoch}.pt')

    def validate_step(self):
        self.model.eval()
        with torch.no_grad():
            total_TD_loss = 0
            total_TD_acc = 0
            for i, data in tqdm(enumerate(self.test_dataLoader), desc="Test", leave=True):
                # 数据预处理
                sentence, labels, distribution = data[0], data[1], data[2]
                sentence = sentence.to(self.device)
                labels = labels.to(self.device)
                distribution = distribution.to(self.device)

                # 模型推理，计算loss
                outputsTD = self.model(sentence)
                TDLoss = self.TimeDistributedLoss(outputsTD, distribution)
                total_TD_loss += TDLoss

                # 计算准确度
                TD_correct = evaluate(outputsTD.clone().detach(), distribution)
                # Label_accuracy = Label_correct*100/labels.numel()
                # TD_accuracy = TD_correct*100/distribution.numel()
                if torch.sum(distribution).item() <= 0.0:
                    TD_accuracy = 100.0
                else:
                    TD_accuracy = TD_correct*100/torch.sum(distribution).item()
                total_TD_acc += TD_accuracy

        print('#Loss:%.5f dacc:%.3f' % (total_TD_loss/len(self.test_dataLoader), total_TD_acc/len(self.test_dataLoader)))
        
        self.test_total_TD_loss = total_TD_loss / len(self.test_dataLoader)
        self.test_total_TD_acc = total_TD_acc / len(self.test_dataLoader)


    def train(self):

        # self.optimizer = optim.Adam(model.parameters(), lr=self.learning_rate)
        self.optimizer = optim.AdamW(model.parameters(), lr=self.learning_rate, weight_decay=0.01)
        # self.scheduler = CosineLRScheduler(optimizer=self.optimizer,
        #                                                     t_initial=self.ditribute_epochs,
        #                                                     lr_min=1e-5,
        #                                                     warmup_t=self.warmup_epochs,
        #                                                     warmup_lr_init=1e-4
        #                                                     )
        self.scheduler = torch.optim.lr_scheduler.ExponentialLR(self.optimizer, gamma=0.95, last_epoch=-1, verbose=True)  

        for epoch in range(self.ditribute_epochs):
            self.train_total_TD_acc = 0
            self.train_total_TD_loss = 0
            self.train_TD_step(epoch=epoch)
            # tensorboard train
            self.writer.add_scalar('Loss/trainTDLoss', self.train_total_TD_loss, epoch)
            self.writer.add_scalar('Acc/trainTDAcc', self.train_total_TD_acc, epoch)
            if epoch>=5 and self.scheduler.get_last_lr()[0]>5e-6:
                self.scheduler.step()
            # print(f'Train Set: lloss:{self.train_total_label_loss}, lacc: {self.train_total_label_acc}')
            print(f'Train Set: tdloss:{self.train_total_TD_loss}, tdacc: {self.train_total_TD_acc}')
            self.test_total_TD_loss = 0
            self.test_total_TD_acc = 0
            self.validate_step()
            # tensorboard validate
            self.writer.add_scalar('Loss/testTDLoss', self.test_total_TD_loss, epoch)
            self.writer.add_scalar('Acc/testTDAcc', self.test_total_TD_acc, epoch)
            print(f'Test Set: tdloss:{self.test_total_TD_loss}, tdacc:{self.test_total_TD_acc}')


def setup_seed(seed):
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    np.random.seed(seed)
    random.seed(seed)

            

if __name__=='__main__':
    parser = argparse.ArgumentParser(
        prog = 'transformerEncoder',
        description='transformerEncoder model to every gene blong to BGC or not',
    )
    parser.add_argument('--datasetPath', required=True)
    parser.add_argument('--max_len', default=64, type=int)
    parser.add_argument('--lmdbPath', required=True)
    # parser.add_argument('--hidden_dim', required=True, type=int)
    parser.add_argument('--nhead', type=int, required=True)
    parser.add_argument('--num_encoder_layers', required=True, default=4, type=int)
    parser.add_argument('--transformer_dropout', default=0.1, type=float)
    parser.add_argument('--mlp_dropout', default=0.5, type=float)
    parser.add_argument('--batch_size', required=True, type=int)
    parser.add_argument('--learning_rate', required=True, type=float)
    parser.add_argument('--interval', required=False, default=10, type=int)
    parser.add_argument('--distribute_epochs', required=True, type=int)
    parser.add_argument('--warmup_epochs', default=200, type=int)
    parser.add_argument('--save_dir', default='./modelSave/')
    parser.add_argument('--load_label_model', action='store_true', required=False)
    parser.add_argument('--label_model_path', required=False)
    parser.add_argument('--two_gpu', required=False, action='store_true')
    parser.add_argument('--seed', required=False, default=42, type=int)
    # parser.add_argument()

    args = parser.parse_args()

    print('--------args----------')
    for k in list(vars(args).keys()):
        print('%s: %s' % (k, vars(args)[k]))
    print('--------args----------\n')
    
    setup_seed(args.seed)
    writer = SummaryWriter('./log/TransformerEncoder/')
    data = DataReader(args.datasetPath, test_ratio=0.2)
    embedding_dim = 320
    if args.load_label_model:
        model = torch.load(args.label_model_path)
    else:
        # model = LSTMNet(embedding_dim=embedding_dim, hidden_dim=embedding_dim, num_layers=args.num_encoder_layers, dropout=args.transformer_dropout)
        model = transformerEncoderNet(d_model=embedding_dim, nhead=args.nhead, num_encoder_layers=args.num_encoder_layers, max_len=args.max_len, 
                                      dim_feedforward=embedding_dim*4, transformer_dropout=args.transformer_dropout, mlp_dropout=args.mlp_dropout, batch_first=True)
        # model = transformerEncoderNet(embedding_dim=embedding_dim, hidden_dim=args.hidden_dim, num_layers=args.num_layers, max_len=args.max_len, labels_num=data.labels_num, dropout=args.dropout)
    if args.two_gpu:
        model = torch.nn.DataParallel(model)
    total = sum(p.numel() for p in model.parameters())
    trainable = sum(p.numel() for p in model.parameters() if p.requires_grad)
    print(f'Total Parameters: {total}\nTrainable Parameters: {trainable}\n')
    TimeDistributedLoss = trainLoss()
    # optimizer = optim.Adam(model.parameters(), lr=args.learning_rate)
    # scheduler = torch.optim.lr_scheduler.ExponentialLR(optimizer, gamma=0.95, last_epoch=-1, verbose=True)  
    trainer = TransformerEncoderTrainer(args=args, writer=writer,data=data, model=model, TimeDistributedLoss=TimeDistributedLoss)
    trainer.train()


import torch
import argparse
import pandas as pd
import numpy as np
from torch.utils.data import DataLoader
import os
from tqdm import tqdm
import pickle
import random
import sys
directory = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.dirname(directory))
# print(sys.path)
from bgc_prophet.train.data import DataReader, BGCLabelsDataset
from bgc_prophet.train.model import transformerEncoderNet
from .baseCommand import baseCommand
from pathlib import Path
import lmdb

class predictCommand(baseCommand):
    name = "predict"
    description = "Predict the gene cluster of given genomes"

    def add_arguments(self, parser):
        parser.add_argument('--datasetPath', type=Path, required=True, help='dataset path')
        parser.add_argument('--modelPath', type=Path, required=True, help='model path')
        parser.add_argument('--outputPath', type=Path, required=False, default=Path('./output/'), help='output path')
        parser.add_argument('--lmdbPath', type=Path, required=True, help='gene representations lmdb path')
        parser.add_argument('--name', type=str, required=False, default='predict', help='name of the output file')
        parser.add_argument('--device', type=str, required=True, choices=["cuda", "cpu"], help='device to use')
        parser.add_argument('--batch_size', type=int, required=False, default=512, help='batch size')
        parser.add_argument('--saveIntermediate', action='store_true', required=False, default=False, help='save intermediate results')

    def handle(self, args):
        predict = genePredicter(args)
        predict.predict()
        if args.saveIntermediate:
            predict.save()

def create_predict_parser():
    parser = argparse.ArgumentParser(
        prog="predict",
        description="Predict the gene cluster of given genomes",
    )
    parser.add_argument('--datasetPath', type=str, required=True, help='dataset path')
    parser.add_argument('--modelPath', type=str, required=True, help='model path')
    parser.add_argument('--outputPath', type=str, required=False, default='./output/', help='output path')
    parser.add_argument('--lmdbPath', type=str, required=True, help='lmdb path')
    parser.add_argument('--name', type=str, required=False, default='predict', help='name of the output file')
    parser.add_argument('--device', type=str, required=True, help='device to use')
    parser.add_argument('--batch_size', type=int, required=False, default=512, help='batch size')
    parser.add_argument('--saveIntermediate', action='store_true', required=False, default=False, help='save intermediate results')
    return parser

class genePredicter:
    def __init__(self, args,):
        self.args = args
        self.datasetPath = args.datasetPath
        self.modelPath = args.modelPath
        self.lmdbPath = args.lmdbPath
        self.outputPath = args.outputPath
        self.device = args.device
        self.batch_size = args.batch_size
        self.name = args.name
        self.saveIntermediate = args.saveIntermediate

        # if not self.saveIntermediate:
        #     self.output = outputFormatter()
        print(self.outputPath)
        os.makedirs(self.outputPath, exist_ok=True)

        self.model = transformerEncoderNet(d_model=320, nhead=5, num_encoder_layers=2, max_len=128,
                                           dim_feedforward=320*4, transformer_dropout=0.1, mlp_dropout=0.5, batch_first=True)
        state_dict = torch.load(self.modelPath, map_location=torch.device('cpu'))
        self.model.load_state_dict(state_dict)
        self.model.to(self.device)
        self.model.eval()

        self.data = DataReader(self.datasetPath, test_ratio=0)
        self.dataset = BGCLabelsDataset(self.data, str(self.lmdbPath), 'eval')
        self.dataLoader = DataLoader(dataset=self.dataset, batch_size=self.batch_size, shuffle=False, num_workers=5)

    def predict(self):
        self.results = []
        for i, data in tqdm(enumerate(self.dataLoader), desc='Predict', leave=True, total=len(self.dataLoader)):
            sentence, labels, distribution = data[0], data[1], data[2]
            sentence = sentence.to(self.device)
            labels = labels.to(self.device)
            distribution = distribution.to(self.device)

            outputsTD = self.model(sentence) 
            outputsTD = outputsTD.cpu().detach().numpy()
            self.results.append(outputsTD)

        self.results = np.concatenate(self.results, axis=0)

    def save(self):
        np.save(self.outputPath.joinpath(f'intermediate_{self.name}.npy'), self.results)

    

# if __name__=='__main__':

#     parser = create_predict_parser()
#     args = parser.parse_args()
#     predict = genePredicter(args)
#     predict.predict()
#     if args.saveIntermediate:
#         predict.save()



#!/usr/bin/env python3 -u
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import argparse
import pathlib
import lmdb
import pickle
import torch
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
import os

from esm import Alphabet, FastaBatchedDataset, ProteinBertModel, pretrained, MSATransformer
from .baseCommand import baseCommand

class ExtractCommand(baseCommand):
    name = "extract"
    description = "Extract all tokens' mean representations and save to a lmdb file for sequences in a FASTA file"

    def add_arguments(self, parser):
        parser.add_argument(
        "model_location",
        type=str,
        help="PyTorch model file OR name of pretrained model to download (see README for models)",
        )
        parser.add_argument(
            "fasta",
            type=pathlib.Path,
            help="FASTA source on which to extract representations",
        )
        parser.add_argument(
            "lmdb_path",
            type=pathlib.Path,
            default=pathlib.Path("./output/lmdb/"),
            help="path to save the lmdb file",
        )
        # parser.add_argument(
        #     "output_dir",
        #     type=pathlib.Path,
        #     help="output directory for extracted representations",
        # )

        parser.add_argument("--toks_per_batch", type=int, default=4096, help="maximum batch size")
        parser.add_argument("--directory", '-d', action='store_true', required=False, default=False, help='indicate the input is a directory')
        parser.add_argument(
            "--repr_layers",
            type=int,
            default=[-1],
            nargs="+",
            help="layers indices from which to extract representations (0 to num_layers, inclusive)",
        )
        parser.add_argument(
            "--include",
            type=str,
            nargs="+",
            choices=["mean", "per_tok", "bos", "contacts"],
            help="specify which representations to return",
            required=True,
        )
        parser.add_argument(
            "--truncation_seq_length",
            type=int,
            default=1022,
            help="truncate sequences longer than the given value",
        )
        parser.add_argument(
            "--write_batches",
            type=int,
            default=1000,
            help="write to lmdb every this many batches, RAM usage increases with this number",
        )

        parser.add_argument("--nogpu", action="store_true", help="Do not use GPU even if available")

    def handle(self, args):
        run(args)



def run(args):
    model, alphabet = pretrained.load_model_and_alphabet(args.model_location)
    model.eval()
    if isinstance(model, MSATransformer):
        raise ValueError(
            "This script currently does not handle models with MSA input (MSA Transformer)."
        )
    if torch.cuda.is_available() and not args.nogpu:
        model = model.cuda()
        print("Transferred model to GPU")

    def extract_and_save(fasta_file, model, alphabet, args):
        args.lmdb_path.mkdir(parents=True, exist_ok=True)
        fasta_seqs = list(SeqIO.parse(fasta_file, "fasta"))
        fasta_id_seqs = [SeqRecord(Seq(str(seq.seq)), id=seq.id, description="") for seq in fasta_seqs]
        temp_file = args.lmdb_path / ('temp_'+str(fasta_file.name))
        SeqIO.write(fasta_id_seqs, temp_file, "fasta")
        dataset = FastaBatchedDataset.from_file(temp_file)
        os.remove(temp_file)
        batches = dataset.get_batch_indices(args.toks_per_batch, extra_toks_per_seq=1)
        data_loader = torch.utils.data.DataLoader(
            dataset, collate_fn=alphabet.get_batch_converter(args.truncation_seq_length), batch_sampler=batches, num_workers=8
        )
        print(f"Read {fasta_file} with {len(dataset)} sequences")

        # args.output_dir.mkdir(parents=True, exist_ok=True)
        map_size = 307374182400
        env = lmdb.open(str(args.lmdb_path), subdir=True, map_size=map_size, readonly=False, meminit=False, map_async=True)
        return_contacts = "contacts" in args.include

        assert all(-(model.num_layers + 1) <= i <= model.num_layers for i in args.repr_layers)
        repr_layers = [(i + model.num_layers + 1) % (model.num_layers + 1) for i in args.repr_layers]

        results = []
        with torch.no_grad():
            for batch_idx, (labels, strs, toks) in enumerate(data_loader):
                print(
                    f"Processing {batch_idx + 1} of {len(batches)} batches ({toks.size(0)} sequences)"
                )
                if torch.cuda.is_available() and not args.nogpu:
                    toks = toks.to(device="cuda", non_blocking=True)

                out = model(toks, repr_layers=repr_layers, return_contacts=return_contacts)

                logits = out["logits"].to(device="cpu")
                representations = {
                    layer: t.to(device="cpu") for layer, t in out["representations"].items()
                }
                if return_contacts:
                    contacts = out["contacts"].to(device="cpu")

                for i, label in enumerate(labels):
                    # args.output_file = args.output_dir / f"{label}.pt"
                    # args.output_file.parent.mkdir(parents=True, exist_ok=True)
                    result = {"label": label}
                    truncate_len = min(args.truncation_seq_length, len(strs[i]))
                    # Call clone on tensors to ensure tensors are not views into a larger representation
                    # See https://github.com/pytorch/pytorch/issues/1995
                    if "per_tok" in args.include:
                        result["representations"] = {
                            layer: t[i, 1 : truncate_len + 1].clone()
                            for layer, t in representations.items()
                        }
                    if "mean" in args.include:
                        result["mean_representations"] = {
                            layer: t[i, 1 : truncate_len + 1].mean(0).clone()
                            for layer, t in representations.items()
                        }
                    if "bos" in args.include:
                        result["bos_representations"] = {
                            layer: t[i, 0].clone() for layer, t in representations.items()
                        }
                    if return_contacts:
                        result["contacts"] = contacts[i, : truncate_len, : truncate_len].clone()

                    results.append(result)

                if batch_idx!=0 and batch_idx%args.write_batches == 0:
                    with env.begin(write=True) as txn:
                        for result_w in results:
                            label_w = result_w["label"]
                            txn.put(label_w.encode('ascii'), pickle.dumps(result_w))
                    results = []
            if results:
                with env.begin(write=True) as txn:
                    for result_w in results:
                        label_w = result_w["label"]
                        txn.put(label_w.encode('ascii'), pickle.dumps(result_w))


    if args.directory:
        if not args.fasta.is_dir():
            raise ValueError("The input is not a directory.")
        else:
            fasta_files = [f for f in args.fasta.iterdir() if f.is_file() and f.suffix in ['.fasta', '.faa']]
            for fasta_file in fasta_files:
                extract_and_save(fasta_file=fasta_file, model=model, alphabet=alphabet, args=args)
    else:
        if args.fasta.is_dir():
            raise ValueError("The input is a directory but not indicated with -d.")
        else:
            fasta_file = args.fasta
            extract_and_save(fasta_file=fasta_file, model=model, alphabet=alphabet, args=args)

# def main():
#     parser = create_parser()
#     args = parser.parse_args()
#     run(args)

# if __name__ == "__main__":
#     main()



import torch
import argparse
import torch.optim as optim
from torch.utils.data import DataLoader
from torch.utils.tensorboard import SummaryWriter   
from tqdm import tqdm
import os
import pickle
import numpy as np
import random

from data import DataReader, BGCLabelsDataset
from loss import trainLoss
from focal_loss import FocalLoss
from utils import evaluate
# from model import transformerEncoderNet
# from classifyTrainer import TransformerClassifier
from classifier import transformerClassifier

class classifyTrainer():
    def __init__(self, args, writer, data, model, loss) -> None:
        self.args = args
        self.writer = writer
        self.data = data
        self.model = model
        self.loss = loss
        self.d_model = data.embedding_dim
        self.lmdbPath = args.lmdbPath
        self.batch_size = args.batch_size
        self.interval = args.interval
        self.epochs = args.epochs
        self.save_dir = args.save_dir
        self.learning_rate = args.learning_rate

        self.train_dataset = BGCLabelsDataset(self.data, self.lmdbPath, 'train')
        print('Train dataset size: ', len(self.train_dataset))
        self.test_dataset = BGCLabelsDataset(self.data, self.lmdbPath, 'test')
        self.train_dataLoader = DataLoader(self.train_dataset, batch_size=self.batch_size, shuffle=True, num_workers=5)
        self.test_dataLoader = DataLoader(self.test_dataset, batch_size=self.batch_size, shuffle=True, num_workers=5)

        print(torch.cuda.is_available())
        self.device = 'cuda' if torch.cuda.is_available() else 'cpu'
        self.model.to(self.device)
        
        self.save_path = self.save_dir + \
        f'transformerClassifier/transformerClassifier_{self.args.max_len}_{self.args.nhead}_{self.args.num_encoder_layers}_{self.args.mlp_dropout}_{self.args.transformer_dropout}_{self.args.learning_rate}_{self.args.epochs}_{self.args.alpha}_{self.args.gamma}/'
        # self.save_path = self.save_dir + \
        # f'transformerClassifier/transformerClassifier_{self.args.max_len}_{self.args.nhead}_ \
        #     {self.args.num_encoder_layers}_{self.args.mlp_dropout}_{self.args.transformer_dropout}_ \
        #     {self.args.learning_rate}_{self.args.epochs}/'
        os.makedirs(os.path.dirname(self.save_path), exist_ok=True)

        print('--------model----------')
        print(self.model)
        print('--------model----------\n')

    def __train_step(self, epoch):
        self.model.train()
        total_loss = 0
        total_acc = 0
        for i, data in tqdm(enumerate(self.train_dataLoader), total=len(self.train_dataLoader), desc='Training', leave=True):
            sentence, labels, distribution = data[0], data[1], data[2]
            sentence = sentence.to(self.device)
            labels = labels.to(self.device)
            distribution = distribution.to(self.device)

            self.optimizer.zero_grad()
            output = self.model(src=sentence, src_key_padding_mask=distribution)
            # print(output.shape, labels.shape)
            loss = self.loss(output, labels)
            total_loss += loss.item()
            correct = evaluate(output.clone().detach(), labels)
            if torch.sum(labels) > 0:
                acc = correct*100/torch.sum(labels).item()
            else:
                acc = 0
            total_acc += acc
            loss.backward()
            self.optimizer.step()

            if i % self.interval == 0:
                print('Epoch: %d, %d/%d, Loss: %.3f, Acc: %.3f' % (epoch, i, len(self.train_dataLoader), total_loss/self.interval, total_acc/self.interval))
                self.train_total_loss += total_loss
                self.train_total_acc += total_acc
                total_loss = 0
                total_acc = 0  
        self.train_total_loss /= len(self.train_dataLoader)
        self.train_total_acc /= len(self.train_dataLoader)
        torch.save(self.model, self.save_path + f'transformerClassifier_{epoch}.pt')
        
    def __val_step(self,):
        self.model.eval()
        with torch.no_grad():
            total_loss = 0
            total_acc = 0
            for i, data in tqdm(enumerate(self.test_dataLoader), total=len(self.test_dataLoader), desc='Testing', leave=True):
                sentence, labels, distribution = data[0], data[1], data[2]
                sentence = sentence.to(self.device)
                labels = labels.to(self.device)
                distribution = distribution.to(self.device)

                output = self.model(src=sentence, src_key_padding_mask=distribution)
                loss = self.loss(output, labels)
                total_loss += loss.item()
                correct = evaluate(output.clone().detach(), labels)

                if torch.sum(labels) > 0:
                    acc = correct*100/torch.sum(labels).item()
                else:
                    acc = 0
                total_acc += acc
            self.test_total_loss = total_loss/len(self.test_dataLoader)
            self.test_total_acc = total_acc/len(self.test_dataLoader)

    def train(self,):
        self.optimizer = optim.AdamW(self.model.parameters(), lr=self.learning_rate, weight_decay=0.01)
        self.scheduler = optim.lr_scheduler.ExponentialLR(self.optimizer, gamma=0.9, last_epoch=-1, verbose=True)

        for epoch in range(self.epochs):
            self.train_total_loss = 0
            self.train_total_acc = 0
            self.__train_step(epoch)
            self.writer.add_scalar('Loss/trainLoss', self.train_total_loss, epoch)
            self.writer.add_scalar('Acc/trainAcc', self.train_total_acc, epoch)
            if epoch>=5 and self.scheduler.get_last_lr()[0]>5e-6:
                self.scheduler.step()
            print(f'Train Set: Epoch: {epoch}, Loss: {self.train_total_loss}, Acc: {self.train_total_acc}')

            self.test_total_loss = 0
            self.test_total_acc = 0
            self.__val_step()
            self.writer.add_scalar('Loss/testLoss', self.test_total_loss, epoch)
            self.writer.add_scalar('Acc/testAcc', self.test_total_acc, epoch)
            print(f'Test Set: Epoch: {epoch}, Loss: {self.test_total_loss}, Acc: {self.test_total_acc}')


def setup_seed(seed):
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)

if __name__=='__main__':
    parser = argparse.ArgumentParser(
        prog='transformerClassifier',
        description='Train a transformer classifier for BGCs',
    )
    parser.add_argument('--datasetPath', type=str, required=True, help='dataset path')
    parser.add_argument('--max_len', type=int, required=True, help='max gene numbers of the input sequence')
    parser.add_argument('--batch_size', type=int, required=True, help='batch size')
    parser.add_argument('--lmdbPath', type=str, required=True, help='training dataset lmdb path')
    parser.add_argument('--nhead', type=int, required=True, help='number of heads in transformer')
    parser.add_argument('--num_encoder_layers', type=int, required=True, help='number of encoder layers in transformer')
    parser.add_argument('--mlp_dropout', type=float, required=False, default=0.5, help='dropout rate of mlp')
    parser.add_argument('--transformer_dropout', type=float, required=False, default=0.1, help='dropout rate of transformer')
    parser.add_argument('--learning_rate', type=float, required=True, help='learning rate')
    parser.add_argument('--interval', type=int, required=False, default=10, help='interval of printing training loss')
    parser.add_argument('--epochs', type=int, required=True, help='epochs')
    parser.add_argument('--seed', type=int, required=False, default=42, help='random seed')
    parser.add_argument('--save_dir', type=str, required=False, default='./modelSave/', help='save dir')
    parser.add_argument('--alpha', type=float, required=False, default=0.25, help='alpha of focal loss')
    parser.add_argument('--gamma', type=float, required=False, default=2, help='gamma of focal loss')

    args = parser.parse_args()
    print('--------args----------')
    for k in list(vars(args).keys()):
        print('%s: %s' % (k, vars(args)[k]))
    print('--------args----------\n')

    setup_seed(args.seed)

    writer = SummaryWriter('./log/transformerClassifier/')
    data = DataReader(args.datasetPath, test_ratio=0.1)
    embedding_dim = 320
    model = transformerClassifier(d_model=embedding_dim, nhead=args.nhead, num_encoder_layers=args.num_encoder_layers, max_len=args.max_len,
                                   dim_feedforward=embedding_dim*4, labels_num=data.labels_num, transformer_dropout=args.transformer_dropout,
                                   mlp_dropout=args.mlp_dropout,)
    total_params = sum(p.numel() for p in model.parameters())
    trainable_params = sum(p.numel() for p in model.parameters() if p.requires_grad)
    print(f'Total params: {total_params}, Trainable params: {trainable_params}')
    # loss = trainLoss()
    loss = FocalLoss(alpha=args.alpha, gamma=args.gamma)
    trainer = classifyTrainer(args, writer, data, model, loss)
    trainer.train()




#!/usr/bin/env python
# -*- coding: UTF-8 -*-

import logging
import os
import argparse
import sys
import torch

from command import ExtractCommand as extract
from command import organizeCommand as organize
from command import splitCommand as split
from command import predictCommand as predict
from command import outputCommand as output
from command import classifyCommand as classify
from command import piplineCommand as pipline


def main():
    print(""" ____   ____  ____      ____                  _          _
| __ ) / ___|/ ___|    |  _ \ _ __ ___  _ __ | |__   ___| |_
|  _ \| |  _| |   _____| |_) | '__/ _ \| '_ \| '_ \ / _ \ __|
| |_) | |_| | |__|_____|  __/| | | (_) | |_) | | | |  __/ |_
|____/ \____|\____|    |_|   |_|  \___/| .__/|_| |_|\___|\__|
                                       |_|
          """)
    parser = argparse.ArgumentParser(prog="BGC-Prophet", description="BGC-Prophet: A tool for BGC mining")
    subparsers = parser.add_subparsers(title="subcommands", 
                                       metavar="COMMAND",
                                       dest='cmd', 
                                       help="Use: bgc_prophet COMMAND -h for more information"
                                       )
    subparsers.required = True
    subcommnds = [
        extract,
        organize,
        split,
        predict,
        output,
        classify,
        pipline,
    ]

    for subcommand in subcommnds:
        cmd_instance = subcommand()
        # print(cmd_instance)
        subparser = subparsers.add_parser(cmd_instance.name, description=cmd_instance.description)
        subparser.set_defaults(handle=cmd_instance.handle)
        cmd_instance.add_arguments(subparser)
    
    args = parser.parse_args()
    if not hasattr(args, 'device'):
        args.device = 'cuda' if torch.cuda.is_available() else 'cpu'
    elif args.device == 'cuda' and not torch.cuda.is_available():
        print('CUDA is not available, use CPU instead')
        args.device = 'cpu'
    if hasattr(args, 'nogpu') and args.nogpu:
        args.device = 'cpu'
        
    if hasattr(args, 'handle'):
        args.handle(args)
    else:
        parser.print_help()


if __name__=='__main__':
    main()
    


import argparse
import torch
import pandas as pd
import numpy as np
from torch.utils.data import DataLoader
import os
from tqdm import tqdm

import sys
from pathlib import Path
# dicrectory = os.path.dirname(os.path.abspath(__file__))
# sys.path.append(os.path.dirname(dicrectory))
from bgc_prophet.train.data import DataReader, BGCLabelsDataset
from bgc_prophet.train.classifier import transformerClassifier
from .baseCommand import baseCommand
import lmdb

class classifyCommand(baseCommand):
    name = "classify"
    description = "Classify the gene cluster of given genomes"

    def add_arguments(self, parser):
        parser.add_argument('--datasetPath', type=Path, required=True, help='dataset path')
        parser.add_argument('--classifierPath', type=Path, required=True, help='classifier path')
        parser.add_argument('--outputPath', type=Path, required=False, default=Path('./output/'), help='output path')
        parser.add_argument('--lmdbPath', type=Path, required=True, help='lmdb path')
        parser.add_argument('--name', type=str, required=False, default='classify', help='name of the output file')
        parser.add_argument('--device', type=str, required=True, help='device to use')
        parser.add_argument('--batch_size', type=int, required=False, default=512, help='batch size')
        parser.add_argument('--classify_t', type=float, required=False, default=0.5, help='threshold for classification')

    def handle(self, args):
        classifier = geneClassifier(args)
        classifier.classify()
        classifier.process()
        classifier.save()

# def create_classify_parser():
#     parser = argparse.ArgumentParser(
#         prog="classify",
#         description="Classify the gene cluster of given genomes",
#     )
#     parser.add_argument('--datasetPath', type=str, required=True, help='dataset path')
#     parser.add_argument('--classifierPath', type=str, required=True, help='classifier path')
#     parser.add_argument('--outputPath', type=str, required=False, default='./output/', help='output path')
#     parser.add_argument('--lmdbPath', type=str, required=True, help='lmdb path')
#     parser.add_argument('--name', type=str, required=False, default='classify', help='name of the output file')
#     parser.add_argument('--device', type=str, required=True, help='device to use')
#     parser.add_argument('--batch_size', type=int, required=False, default=512, help='batch size')
#     parser.add_argument('--classify_t', type=float, required=False, default=0.5, help='threshold for classification')

#     return parser

class geneClassifier:
    def __init__(self, args) -> None:
        self.args = args
        self.datasetPath = args.datasetPath
        self.classifierPath = args.classifierPath
        self.lmdbPath = args.lmdbPath
        self.outputPath = args.outputPath
        self.device = args.device
        self.batch_size = args.batch_size
        self.name = args.name
        self.classify_t = args.classify_t

        self.model = transformerClassifier(d_model=320, nhead=5, num_encoder_layers=2, max_len=128, 
                                           dim_feedforward=320*4, labels_num=7, transformer_dropout=0.1, 
                                           mlp_dropout=0.5)
        state_dict = torch.load(self.classifierPath, map_location=torch.device('cpu'))
        self.model.load_state_dict(state_dict)
        self.model.to(self.device)
        self.model.eval()
        # print(self.datasetPath)
        self.data = DataReader(self.datasetPath, test_ratio=0)
        self.dataset = BGCLabelsDataset(self.data, str(self.lmdbPath), 'eval')
        self.dataLoader = DataLoader(dataset=self.dataset, batch_size=self.batch_size, shuffle=False, num_workers=5)
        self.OtherIndex = self.data.labels_list.index('Other')
        
    def classify(self):
        self.results = []
        for i, data in tqdm(enumerate(self.dataLoader), desc='Classify', leave=True, total=len(self.dataLoader)):
            sentence, labels, distribution = data[0], data[1], data[2]
            sentence = sentence.to(self.device)
            labels = labels.to(self.device)
            distribution = distribution.to(self.device)

            output = self.model(src=sentence, src_key_padding_mask=distribution)
            classes = output.detach().cpu().numpy()
            self.results.append(classes)
        self.results = np.concatenate(self.results, axis=0)

    def process(self):
        self.classes = self.results.copy()
        dataFrame = pd.read_csv(self.datasetPath)
        labels_list = self.data.labels_list
        for i, row in dataFrame.iterrows():
            if row['isBGC'] == 'No':
                dataFrame.loc[i, 'labels'] = 'NonBGC'
                dataFrame.loc[i, 'probability'] = 1.
                continue
            result = self.results[i]
            prediction = result.copy()
            threshold = max(result[self.OtherIndex], self.classify_t)
            prediction[result>=threshold] = 1
            prediction[result<threshold] = 0
            if 1 in prediction:
                labels = [labels_list[j] for j in range(len(labels_list)) if prediction[j]==1]
                probability = [result[j] for j in range(len(labels_list)) if prediction[j]==1]
                if 'Other' in labels and len(labels)>1:
                    oindex = labels.index('Other')
                    labels.pop(oindex)
                    probability.pop(oindex)
                dataFrame.loc[i, 'labels'] = ' '.join(labels)
                dataFrame.loc[i, 'probability'] = sum(probability)/len(probability)
            else:
                dataFrame.loc[i, 'labels'] = 'Unknown'
                dataFrame.loc[i, 'probability'] = 1 - np.max(result)
            
        self.dataFrame = dataFrame

    def save(self):
        np.save(self.outputPath.joinpath(self.name+'_results.npy'), self.results)
        self.dataFrame['probability'] = self.dataFrame['probability'].apply(lambda x: '%.3f' % x)
        self.dataFrame.to_csv(self.outputPath.joinpath(self.name+'_classified.csv'), index=False)

# if __name__ == "__main__":
#     parser = create_classify_parser()
#     args = parser.parse_args()
#     classifier = geneClassifier(args)
#     classifier.classify()
#     classifier.process()
#     classifier.save()
